@@ -232,6 +232,8 @@ create_homedir(const struct passwd *pwd,
       {
          pam_syslog(NULL, LOG_DEBUG,
 		    "unable to open or stat src file %s: %m", newsource);
+         if (srcfd >= 0)
+            close(srcfd);
          closedir(d);
 
 #ifndef PATH_MAX
@@ -373,7 +375,7 @@ main(int argc, char *argv[])
    pwd = getpwnam(argv[1]);
    if (pwd == NULL) {
 	pam_syslog(NULL, LOG_ERR, "User unknown.");
-	return PAM_CRED_INSUFFICIENT;
+	return PAM_USER_UNKNOWN;
    }
 
    if (argc >= 3) {
